/// @author: iMujtaba Nazki
/// Date: 03-04-2019
/// Copyright: im8488 @: 2019

import 'package:flutter/material.dart';

/// A BrickListItem is created on the analogy of a wall and bricks. The container
/// that holds all the widgets is called a 'wall' and the individual widgets
/// within the wall are called as 'bricks'. When bricks are added to the wall they
/// automatically resize and position themselves to fit the wall. It also allows
/// for nesting, i.e. you can nest BrickListItems inside another BrickListItem.
/// [_widgets] are the bricks to be added to the wall, while as [_wallColor] and
/// [_brickColor] specify the color of the wall and the widgets i.e. bricks
/// within the wall, respectively. You can wrap BrickListItem within the
/// corresponding widget which handles user interaction, to support interaction
/// for the BrickListItem.
class BrickListItem extends StatelessWidget {
  final List<Widget> _widgets;
  final Color _wallColor;
  final Color _brickColor;

  /// Constructs a BrickListItem with default settings.
  BrickListItem({
    List<Widget> widgets,
    Color wallColor = Colors.black,
    Color brickColor = Colors.red,
  })  : _widgets = widgets,
        _wallColor = wallColor,
        _brickColor = brickColor;

  /// Constructs a BrickListItem with black [_wallColor] and red [_brickColor].
  BrickListItem.home({
    List<Widget> widgets,
    Color wallColor = Colors.black,
    Color brickColor = Colors.red,
  })  : _widgets = widgets,
        _wallColor = wallColor,
        _brickColor = brickColor;

  /// Constructs a BrickListItem with green [_wallColor] and brown [_brickColor].
  BrickListItem.mud({List<Widget> widgets})
      : _widgets = widgets,
        _wallColor = Colors.green,
        _brickColor = Colors.brown;

  /// Constructs a BrickListItem with black [_wallColor] and brown [_brickColor].
  BrickListItem.forest({List<Widget> widgets})
      : _widgets = widgets,
        _wallColor = Colors.black,
        _brickColor = Colors.brown;

  /// Constructs a BrickListItem with blue [_wallColor] and white [_brickColor].
  BrickListItem.ocean({List<Widget> widgets})
      : _widgets = widgets,
        _wallColor = Colors.blue,
        _brickColor = Colors.white;

  /// Constructs a BrickListItem with lightBlue [_wallColor] and lightBlueAccent
  /// [_brickColor].
  BrickListItem.sky({List<Widget> widgets})
      : _widgets = widgets,
        _wallColor = Colors.lightBlue,
        _brickColor = Colors.lightBlueAccent;

  /// Constructs a BrickListItem with transparent [_wallColor] and [_brickColor].
  BrickListItem.transparent({List<Widget> widgets})
      : _widgets = widgets,
        _wallColor = Colors.transparent,
        _brickColor = Colors.transparent;

  @override
  Widget build(BuildContext context) {
    return _wall();
  }

  /// A wall is a container within a BrickListItem that holds widgets added to
  /// the BrickListItem. In terms of customization, it color can be changed. It
  /// is responsible for automatically positioning and sizing the widgets added
  /// to the BrickListItem. By default, a wall has a margin of 2px and padding
  /// of 1px on all sides, along with a borderRadius of 2px, which at the moment
  /// are not customizable.
  Widget _wall() {
    return FittedBox(
      fit: BoxFit.contain,
      child: Container(
        margin: const EdgeInsets.only(
          left: 6.0,
          top: 4.0,
          right: 6.0,
        ),
        padding: const EdgeInsets.symmetric(horizontal: 1.0, vertical: 1.0),
        decoration: BoxDecoration(
          color: _wallColor,
          borderRadius: BorderRadius.circular(2.0),
        ),
        child: Row(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: _bricks(),
        ),
      ),
    );
  }

  /// When widgets are added to the BrickListItem, they are first converted to
  /// bricks, which are later added to the wall. All bricks have a margin of 2px,
  /// horizontal padding of 5px and vertical padding of 8px along with a borderRadius
  /// of 2px, which at the moment are not customizable.
  List<Widget> _bricks() {
    List<Widget> bricks = [];
    _widgets.forEach(
      (widget) {
        Container brick = Container(
          margin: const EdgeInsets.all(2.0),
          padding: const EdgeInsets.symmetric(horizontal: 5.0, vertical: 8.0),
          decoration: BoxDecoration(
            color: _brickColor,
            borderRadius: BorderRadius.circular(2.0),
          ),
          child: widget,
        );
        bricks.add(brick);
      },
    );
    return bricks;
  }
}
