import 'package:flutter/material.dart';

/// Implementation that must be provided by the user, for what should happen
/// when a FlatListItem is tapped. Although, a default implementation is provided
/// which simply prints the message, that a user implementation is required.
typedef OnItemTap = Function({BuildContext context, int tappedAt});

/// A set of predefined presets to control the size of the FlatListItem. Based
/// on the preset selected, the height of the holder and the size of the items
/// it contains is scaled accordingly. A preset calculates the height of the holder
/// based on the device orientation and divided by corresponding factor. Following
/// are the factors used in each preset and depending on the device orientation:-
/// [minimal] : 14.0 (P) and 7.0 (L), [small] : 10.0 (P) and 7.0 (L), [normal] :
/// 9.0 (P) and 6.0 (L), [large] 6.0 (P) and 4.0 (L), and finally [gigantic] :
/// 3.0 (P/L).
enum ScalePreset {
  minimal,
  small,
  normal,
  large,
  gigantic,
}

/// A FlatListItem is typically added to a ListView to display textual information.
/// FlatListItem is divided into three columns, wherein, the first one on the left
/// displays an icon, second and the third column each have 2 rows, wherein,
/// the first row is typically used to display the headings to represent the
/// data that is displayed in the second row. The container that holds all these
/// columns and items is called as a 'holder' in a FlatListItem, whose dimensions
/// are calculated automatically based on the dimensions of the user device
/// screen and the [ScalePreset] chosen by the user. Also, based on the [ScalePreset]
/// chosen, all the items of the holder are scaled accordingly. It is important
/// to know that if a FlatListItem is added to a ListView, the [indexInList] field
/// is typically used to hold its index in the ListView to which it is added, so
/// that it becomes possible to query a FlatListItem when added to ListView to
/// get its position within the ListView.
class FlatListItem extends StatelessWidget {
  final BuildContext _context;
  final int _indexInList;
  final ScalePreset _scalePreset;
  final double _cornerRadius;
  final Icon _icon;
  final String _leftLabelText;
  final String _leftDataText;
  final String _rightLabelText;
  final String _rightDataText;
  final Color _backgroundColor;
  final Color _leftLabelTextColor;
  final Color _leftDataTextColor;
  final Color _rightLabelTextColor;
  final Color _rightDataTextColor;
  final Color _hangingDividerColor;
  final FontWeight _leftLabelTextFontWeight;
  final FontWeight _leftDataTextFontWeight;
  final FontWeight _rightLabelTextFontWeight;
  final FontWeight _rightDataTextFontWeight;
  final bool _showHangingDivider;
  final OnItemTap _onTap;

  /// Constructs a FlatListItem with white background. This is the primary constructor
  /// that is also the most customizable.
  FlatListItem({
    @required BuildContext context,
    int indexInList = 0,
    double cornerRadius = 0.0,
    ScalePreset scalePreset = ScalePreset.normal,
    Color backgroundColor = Colors.white,
    Icon icon,
    String leftLabelText,
    String leftDataText,
    String rightLabelText,
    String rightDataText,
    Color leftLabelTextColor = Colors.grey,
    Color leftDataTextColor = Colors.black,
    Color rightLabelTextColor = Colors.grey,
    Color rightDataTextColor = Colors.black,
    FontWeight leftLabelTextFontWeight = FontWeight.normal,
    FontWeight leftDataTextFontWeight = FontWeight.normal,
    FontWeight rightLabelTextFontWeight = FontWeight.normal,
    FontWeight rightDataTextFontWeight = FontWeight.normal,
    bool showHangingDivider = true,
    Color hangingDividerColor = Colors.lightGreen,
    OnItemTap onTap,
  })  : _context = context,
        _indexInList = indexInList,
        _scalePreset = scalePreset,
        _cornerRadius = cornerRadius,
        _backgroundColor = backgroundColor,
        _icon = icon,
        _leftLabelText = leftLabelText,
        _leftDataText = leftDataText,
        _rightLabelText = rightLabelText,
        _rightDataText = rightDataText,
        _leftLabelTextColor = leftLabelTextColor,
        _leftDataTextColor = leftDataTextColor,
        _rightLabelTextColor = rightLabelTextColor,
        _rightDataTextColor = rightDataTextColor,
        _hangingDividerColor = hangingDividerColor,
        _leftLabelTextFontWeight = leftLabelTextFontWeight,
        _leftDataTextFontWeight = leftDataTextFontWeight,
        _rightLabelTextFontWeight = rightLabelTextFontWeight,
        _rightDataTextFontWeight = rightDataTextFontWeight,
        _showHangingDivider = showHangingDivider,
        _onTap = onTap;

  /// Constructs a FlatListItem with dark background.
  FlatListItem.dark({
    @required BuildContext context,
    int indexInList = 0,
    double cornerRadius = 0.0,
    ScalePreset scalePreset = ScalePreset.normal,
    Icon icon,
    String leftLabelText,
    String leftDataText,
    String rightLabelText,
    String rightDataText,
    FontWeight centerTitleFontWeight = FontWeight.normal,
    FontWeight centerSubTitleFontWeight = FontWeight.normal,
    FontWeight trailingTitleFontWeight = FontWeight.normal,
    FontWeight trailingSubTitleFontWeight = FontWeight.normal,
    bool showHangingDivider = true,
    OnItemTap onTap,
  }) : this(
          context: context,
          indexInList: indexInList,
          cornerRadius: cornerRadius,
          scalePreset: scalePreset,
          icon: icon,
          leftLabelText: leftLabelText,
          leftDataText: leftDataText,
          rightLabelText: rightLabelText,
          rightDataText: rightDataText,
          leftLabelTextFontWeight: centerTitleFontWeight,
          leftDataTextFontWeight: centerSubTitleFontWeight,
          rightLabelTextFontWeight: trailingTitleFontWeight,
          rightDataTextFontWeight: trailingSubTitleFontWeight,
          showHangingDivider: showHangingDivider,
          onTap: onTap,

          //Pre-configured.
          backgroundColor: Colors.black,
          leftLabelTextColor: Colors.yellow,
          leftDataTextColor: Colors.white,
          rightLabelTextColor: Colors.yellow,
          rightDataTextColor: Colors.white,
          hangingDividerColor: Colors.white,
        );

  /// Constructs a FlatListItem with purple background.
  FlatListItem.purple({
    @required BuildContext context,
    int indexInList = 0,
    double cornerRadius = 0.0,
    ScalePreset scalePreset = ScalePreset.normal,
    Icon icon,
    String leftLabelText,
    String leftDataText,
    String rightLabelText,
    String rightDataText,
    FontWeight leftLabelTextFontWeight = FontWeight.normal,
    FontWeight leftDataTextFontWeight = FontWeight.normal,
    FontWeight rightLabelTextFontWeight = FontWeight.normal,
    FontWeight rightDataTextFontWeight = FontWeight.normal,
    bool showHangingDivider = true,
    OnItemTap onTap,
  }) : this(
          context: context,
          indexInList: indexInList,
          cornerRadius: cornerRadius,
          scalePreset: scalePreset,
          icon: icon,
          leftLabelText: leftLabelText,
          leftDataText: leftDataText,
          rightLabelText: rightLabelText,
          rightDataText: rightDataText,
          leftLabelTextFontWeight: leftLabelTextFontWeight,
          leftDataTextFontWeight: leftDataTextFontWeight,
          rightLabelTextFontWeight: rightLabelTextFontWeight,
          rightDataTextFontWeight: rightDataTextFontWeight,
          showHangingDivider: showHangingDivider,
          onTap: onTap,

          //Pre-configured field.
          backgroundColor: Colors.deepPurple,
          leftLabelTextColor: Colors.white,
          leftDataTextColor: Colors.yellow,
          rightLabelTextColor: Colors.white,
          rightDataTextColor: Colors.yellow,
          hangingDividerColor: Colors.yellow,
        );

  /// Constructs a FlatListItem with transparent background.
  FlatListItem.transparent({
    @required BuildContext context,
    int indexInList = 0,
    double cornerRadius = 0.0,
    ScalePreset scalePreset = ScalePreset.normal,
    Icon icon,
    String leftLabelText,
    String leftDataText,
    String rightLabelText,
    String rightDataText,
    FontWeight leftLabelTextFontWeight = FontWeight.normal,
    FontWeight leftDataTextFontWeight = FontWeight.normal,
    FontWeight rightLabelTextFontWeight = FontWeight.normal,
    FontWeight rightDataTextFontWeight = FontWeight.normal,
    bool showHangingDivider = true,
    OnItemTap onTap,
  }) : this(
          context: context,
          indexInList: indexInList,
          cornerRadius: cornerRadius,
          scalePreset: scalePreset,
          icon: icon,
          leftLabelText: leftLabelText,
          leftDataText: leftDataText,
          rightLabelText: rightLabelText,
          rightDataText: rightDataText,
          leftLabelTextFontWeight: leftLabelTextFontWeight,
          leftDataTextFontWeight: leftDataTextFontWeight,
          rightLabelTextFontWeight: rightLabelTextFontWeight,
          rightDataTextFontWeight: rightDataTextFontWeight,
          showHangingDivider: showHangingDivider,
          onTap: onTap,

          //Pre-configured fields.
          backgroundColor: Colors.transparent,
          leftLabelTextColor: Colors.green,
          leftDataTextColor: Colors.black,
          rightLabelTextColor: Colors.green,
          rightDataTextColor: Colors.black,
          hangingDividerColor: Colors.grey,
        );

  /// Constructs a FlatListItem with light-green background.
  FlatListItem.lightGreen({
    @required BuildContext context,
    int indexInList = 0,
    double cornerRadius = 0.0,
    ScalePreset scalePreset = ScalePreset.normal,
    Icon icon,
    String leftLabelText,
    String leftDataText,
    String rightLabelText,
    String rightDataText,
    FontWeight leftLabelTextFontWeight = FontWeight.normal,
    FontWeight leftDataTextFontWeight = FontWeight.normal,
    FontWeight rightLabelTextFontWeight = FontWeight.normal,
    FontWeight rightDataTextFontWeight = FontWeight.normal,
    bool showHangingDivider = true,
    OnItemTap onTap,
  }) : this(
          context: context,
          indexInList: indexInList,
          cornerRadius: cornerRadius,
          scalePreset: scalePreset,
          icon: icon,
          leftLabelText: leftLabelText,
          leftDataText: leftDataText,
          rightLabelText: rightLabelText,
          rightDataText: rightDataText,
          leftLabelTextFontWeight: leftLabelTextFontWeight,
          leftDataTextFontWeight: leftDataTextFontWeight,
          rightLabelTextFontWeight: rightLabelTextFontWeight,
          rightDataTextFontWeight: rightDataTextFontWeight,
          showHangingDivider: showHangingDivider,
          onTap: onTap,

          //Pre-configured fields. Color = LightGreen.
          backgroundColor: Color(0xAEAED581),
          leftLabelTextColor: Colors.blue,
          leftDataTextColor: Colors.black,
          rightLabelTextColor: Colors.blue,
          rightDataTextColor: Colors.black,
          hangingDividerColor: Colors.black,
        );

  @override
  Widget build(BuildContext context) {
    return _holder();
  }

  /// Holder is the primary container which holds every other item in the
  /// FlatListItem. Its width is computed based on the width of the device screen
  /// minus horizontal padding of '20px' i.e. '10px' on left and '10px' on right.
  /// It also has a vertical padding of '5px'. When a holder is tapped, it also
  /// displays an InkWell animation.
  Widget _holder() {
    return InkWell(
      onTap: () => _onTap == null
          ? _defaultInteractImplementation()
          : _onTap(context: _context, tappedAt: _indexInList),
      child: Container(
        // Margins around the holder.
        margin: const EdgeInsets.symmetric(
          horizontal: 10.0,
          vertical: 5.0,
        ),
        height: _scalePresetInDouble(_scalePreset),
        width: MediaQuery.of(_context).size.width,
        decoration: BoxDecoration(
          color: _backgroundColor,
          borderRadius: BorderRadius.circular(_cornerRadius),
        ),
        child: Row(
          children: <Widget>[
            _icon != null ? _leadingIcon() : Container(),
            _leftLabelText != null && _leftDataText != null
                ? _leftColumn()
                : Container(),
            _showHangingDivider
                ? _leftLabelText != null && _leftDataText != null
                    ? _hangingVerticalDivider()
                    : Container()
                : Container(),
            _rightColumn(),
          ],
        ),
      ),
    );
  }

  /// The leadingIcon is simply and icon and is the first thing that appears in
  /// a FlatListItem, only if it is set to visible. Review:
  Widget _leadingIcon() {
    return Container(
      width: MediaQuery.of(_context).size.width / 10.0,
      child: Align(
        alignment: Alignment.center,
        child: _icon,
      ),
    );
  }

  /// The left-column which contains two rows i.e. 'top' and 'bottom', wherein,
  /// the top row contains the labelText and the bottom row contains the dataText.
  /// Typically used to present the outline of the data to the user. Its width
  /// is calculated by dividing the user's deviceScreen width by a factor of
  /// '4.0' in portrait mode and by a factor of '6.0' when the user's device is
  /// in landscape mode. Review:
  Widget _leftColumn() {
    final deviceOrientation = MediaQuery.of(_context).orientation;
    final widthInPortraitMode = MediaQuery.of(_context).size.width / 4.0;
    final widthInLandscapeMode = MediaQuery.of(_context).size.width / 6.0;

    return Container(
      width: _leftLabelText != null && _leftDataText != null
          ? deviceOrientation == Orientation.portrait
              ? widthInPortraitMode
              : widthInLandscapeMode
          : 0.0,
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.center,
        mainAxisAlignment: MainAxisAlignment.center,
        children: <Widget>[
          _leftLabelText != null
              ? Text(
                  _leftLabelText,
                  style: TextStyle(
                    fontSize: _computedLabelTextFontSize(),
                    color: _leftLabelTextColor,
                    fontWeight: _leftLabelTextFontWeight,
                  ),
                  maxLines: 1,
                  overflow: TextOverflow.ellipsis,
                )
              : Container(),
          SizedBox(
            height: 4.0,
          ),
          _leftDataText != null
              ? Text(
                  _leftDataText,
                  style: TextStyle(
                    fontSize: _computedDataTextFontSize(),
                    color: _leftDataTextColor,
                    fontWeight: _leftDataTextFontWeight,
                  ),
                  maxLines: computedMaxLinesToDisplay(),
                  overflow: TextOverflow.ellipsis,
                )
              : Container()
        ],
      ),
    );
  }

  /// The right-most column which contains two rows i.e. 'top' and 'bottom',
  /// wherein, the top row contains the labelText and the bottom row contains
  /// the dataText. Typically used to present the description of the data that
  /// is specified in the left-column. A padding of '10.0px' is applied on both
  /// sides of the right-column. Also, the text within the rightDataText row
  /// is 'justified' to give it a pleasing appearance. Review:
  Widget _rightColumn() {
    // Trailing takes the rest of the space.
    return Expanded(
      child: Padding(
        padding: const EdgeInsets.only(left: 10.0, right: 10.0),
        child: Column(
          crossAxisAlignment: CrossAxisAlignment.start,
          mainAxisAlignment: MainAxisAlignment.center,
          children: <Widget>[
            _rightLabelText != null
                ? Text(
                    _rightLabelText,
                    style: TextStyle(
                      fontSize: _computedLabelTextFontSize(),
                      color: _rightLabelTextColor,
                      fontWeight: _rightLabelTextFontWeight,
                    ),
                    maxLines: 1,
                    overflow: TextOverflow.ellipsis,
                  )
                : Container(
                    width: 5.0,
                    height: 5.0,
                  ),
            SizedBox(
              height: 4.0,
            ),
            _rightDataText != null
                ? Text(
                    _rightDataText,
                    style: TextStyle(
                      fontSize: _computedDataTextFontSize(),
                      color: _rightDataTextColor,
                      fontWeight: _rightDataTextFontWeight,
                    ),
                    softWrap: true,
                    maxLines: computedMaxLinesToDisplay(),
                    overflow: TextOverflow.ellipsis,
                    textAlign: TextAlign.justify,
                  )
                : Container()
          ],
        ),
      ),
    );
  }

  /// Returns a customized VerticalDivider that is padded on the top by 15.0px
  /// and padded on the bottom by 15.0px. Padding along the top and the bottom
  /// gives a hanging effect to the VerticalDivider, hence the name.
  Widget _hangingVerticalDivider() {
    return Padding(
      padding: EdgeInsets.only(
        top: 15,
        bottom: 15,
      ),
      child: VerticalDivider(
        color: _hangingDividerColor,
      ),
    );
  }

  /// Returns the height of the holder in double. Typically used internally to
  /// calculate the height of the holder based on the heightPreset chosen by the
  /// class-user. [minimal] when the heigh is divided by a factor of 14 when in
  /// portrait and by 8 when in landscape mode.
  double _scalePresetInDouble(ScalePreset customHeight) {
    switch (customHeight) {
      case ScalePreset.minimal:
        return _computedHolderHeight(14, 8);

      case ScalePreset.small:
        return _computedHolderHeight(10, 7);

      case ScalePreset.large:
        return _computedHolderHeight(6, 4);

      case ScalePreset.gigantic:
        return _computedHolderHeight(3, 3);

      // Default is ScalePreset.normal
      default:
        return _computedHolderHeight(9, 6);
    }
  }

  /// Returns the height of the holder by dividing the height of the device
  /// either by the [portraitFactor] or by the [landscapeFactor] depending
  /// on the device orientation. A factor determines the times the screen
  /// height should be divided, and thats what is returned based on the device
  /// orientation.
  double _computedHolderHeight(double portraitFactor, double landscapeFactor) {
    Orientation deviceOrientation = MediaQuery.of(_context).orientation;
    double deviceHeight = MediaQuery.of(_context).size.height;

    return deviceOrientation == Orientation.portrait
        ? deviceHeight / portraitFactor
        : deviceHeight / landscapeFactor;
  }

  /// The term 'computed' is used because the fontSize is returned based on the
  /// [ScalePreset] specified by the user. That is, a specific label size is
  /// returned based on what [ScalePreset] is specified by the user. Review:
  double _computedLabelTextFontSize() {
    switch (_scalePreset) {
      case ScalePreset.minimal:
        return 10.0;

      case ScalePreset.small:
        return 11.0;

      case ScalePreset.large:
        return 13.0;

      case ScalePreset.gigantic:
        return 14.0;

      // ScalePresent.normal
      default:
        return 12.0;
    }
  }

  /// The term 'computed' is used because the fontSize is returned based on the
  /// [ScalePreset] specified by the user. That is, a specific label size is
  /// returned based on what [ScalePreset] is specified by the user. Review:
  double _computedDataTextFontSize() {
    switch (_scalePreset) {
      case ScalePreset.minimal:
        return 11.0;

      case ScalePreset.small:
        return 13.0;

      case ScalePreset.large:
        return 15.0;

      case ScalePreset.gigantic:
        return 16.0;

      // ScalePresent.normal
      default:
        return 14.0;
    }
  }

  /// Returns the maximum lines to display in the rightDataText row based on the
  /// ScalePreset chosen by the user. The smaller the ScalePreset chosen the
  /// lesser the number of lines displayed and vice-versa. When [ScalePreset.minimal]
  /// is selected '1' line is displayed, '2' lines in case of [ScalePreset.small]
  /// and [ScalePreset.normal], '3' in case of [ScalePreset.large] and '8' in
  /// Portrait orientation and '5' in Landscape orientation for [ScalePreset.gigantic].
  int computedMaxLinesToDisplay() {
    Orientation deviceOrienation = MediaQuery.of(_context).orientation;

    switch (_scalePreset) {
      case ScalePreset.minimal:
        return 1;

      case ScalePreset.large:
        return 3;

      case ScalePreset.gigantic:
        return deviceOrienation == Orientation.portrait ? 8 : 5;

      // ScalePreset.normal and ScalePreset.small
      default:
        return 2;
    }
  }

  /// Default implementation for what happens when the user interacts with the
  /// FlatListItem and no implementation is provided for that event.
  void _defaultInteractImplementation() =>
      print('No implementation. Please provide one!');
}
